/********************************************************************/
/*                                                                  */
/* Program name: MQWSEC.C                                           */
/*                                                                  */
/* Description:  Source file for the Sample Security Exit DLL       */
/*                                                                  */
/*  Details:      Built into the Security Exit is a process which   */
/*                simply allows for conditions where the same IP    */
/*                address can be allocated to several machines.     */
/*                Such as when DHCP or serial link connection       */
/*                protocols such as SLIP /PPP where the Host        */
/*                allocates IP address each time a call is made.    */
/*                                                                  */
/*                This Security Exit is used to overcome the        */
/*                current limitations of MQ which uses the IP       */
/*                address as part of the key to used to synchronsise*/
/*                messages/data for each MQ client.                 */
/*                This should not be confused with the MessageID    */
/*                that is passed with each message header.          */
/*                Later releases may address this problem thus      */
/*                making the use of the Channel Security Exit       */
/*                redundant for this...                             */
/*                                                                  */
/*  Security &                                                      */
/*  the fix:                                                        */
/*                We have maintained the possibility of using the   */
/*                Security Exit as a Security Exit, as well as      */
/*                passing a unique machine identifier or Qmanager   */
/*                name from the Client to the Host as part of the   */
/*                response from the Client, by appending it to      */
/*                the end of the response.                          */
/*                                                                  */
/* Fix and no                                                       */
/* Security data:                                                   */
/*                In the event NO security is required then simply  */
/*                leaving the  security exit data field  blank will */
/*                do the trick. (Make sure there are NO blanks in   */
/*                the empty Security Data field).                   */
/*                                                                  */
/*                Although the only Channel exit required is the    */
/*                security exit, we have included  entry points     */
/*                for MessageExit()                                 */
/*                    SendExit()                                    */
/*                    ReceiveExit() that may be of use in developing*/
/*                application.                                      */
/*                                                                  */
/*  NOTE:         This sample must be run at both the Client and the*/
/*                Host. Athough for other platforms some changes    */
/*                will be nescessary.                               */
/*                                                                  */
/*                This exit is NOT required at all if unique IP     */
/*                addresses can be guaranteed.                      */
/*------------------------------------------------------------------*/
/*  TRACE:        There is a built in trace facility which shows    */
/*                what data is passed during the exits etc.         */
/*                and are there only to assist in the use or debug  */
/*                of your exits, the trace file 'sectrace.log'      */
/*                is placed in \mqw\data directory                  */
/*                                                                  */
/*               To turn the trace on set environment variable      */
/*               MQWXTRACE=YES in \autoexec.bat before starting MQW */
/*------------------------------------------------------------------*/
/*                                                                  */
/*  Disclaimer:                                                     */
/*                This is a working sample, with extra frills such  */
/*                as trace. It is up to the application developer   */
/*                to decide on the actual implementation as long    */
/*                as a uniqe ID is SET in the 'ConnectionName'      */
/*                parameter of the Channel Descriptor at the remote */
/*                end.                                              */
/*                                                                  */
/********************************************************************/
/*                                                                  */
/* Statement:     Licensed Materials - Property of IBM              */
/*                                                                  */
/*                5622-960 (C) Copyright IBM Corp. 1996             */
/*                                                                  */
/*                U.S. Government Users Restricted Rights - use,    */
/*                duplication or disclosure restricted by GSA ADP   */
/*                Schedule Contract with IBM Corp.                  */
/*                                                                  */
/********************************************************************/

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <cmqc.h>
#include <cmqxc.h>
#include <process.h>
#include "mqwsec.h"

/* OpenTraceFile: Subroutine to open output trace file */

//  This define increases the code generated but speeds up the execution !
//  Alternatively remove the trace altogether when you are happy with your code
//
#define Traceit if(dotrace) Trace


//#define CLOSE_TRACE 1         // remove the comment if your program crashes
                                // and you don't get all the trace you expect


// -------------------- STATIC VARIABLES---------------------------------

BOOL dotrace=FALSE;
char *pszValue;
char szChltraceFilename[260];
BOOL TraceFileOpened = FALSE;
FILE *stream = NULL;

// --------------------  Tracing code   ---------------------------------

void OpenTraceFile( void )
{
   time_t ltime;

 pszValue = getenv("MQWXTRACE");

 if( (pszValue !=NULL) &&  !strcmp(pszValue,"YES" ) )
  {
   dotrace=TRUE;
   if (TraceFileOpened)              // Open for append if opened before
   {
     stream = fopen( szChltraceFilename, "a" );
   }
   else
   {
    GetProfileString( "MQWIN",
                      "DATADIR",
                      "\\MQW\\DATA\\",
                      szChltraceFilename,
                      sizeof(szChltraceFilename));

    strcat(szChltraceFilename, "sectrace.log");
    stream = fopen( szChltraceFilename, "w" );
    TraceFileOpened = TRUE;
    time(&ltime);
    fprintf(stream, "Channel trace started on %s\n", ctime( &ltime ));
   }
  }
  else
  {
   dotrace=FALSE;
  }
}


void
Trace(char *msg,...)
{

 if(dotrace)
 {
 va_list argptr;

#ifdef CLOSE_TRACE              // closing trace file every write saves data!
                                // but slows the system down.
                                // use to find bad bugs !
   if (stream==NULL)
   {
     stream = fopen( szChltraceFilename, "a" );
   }
   // what if open fails ! , check NULL stream
#endif

   va_start(argptr, msg);
   vfprintf(stream,msg,argptr);
   va_end(argptr);


#ifdef CLOSE_TRACE
   if(stream !=NULL) fclose(stream);
   stream=NULL;
#endif

 }
}


//
// -------------------- Check and trace exit parameters   ---------------
//

void CheckMQCXP (MQCXP *pChannelExitParms)
{
   int i;
   unsigned char *p;

  /* Check struc id */
  if (strncmp(pChannelExitParms->StrucId,
              MQCXP_STRUC_ID,
              sizeof(MQCHAR4)
             )
     )
  {
     Traceit( "  *** Found incorrect MQCXP Struc id: %.s\n",
             pChannelExitParms->StrucId);
  }

  /* Check version */

  if (pChannelExitParms->Version != MQCXP_VERSION_2) {
     Traceit( "  *** Found incorrect MQCXP Version: %ld\n",
             pChannelExitParms->Version);
  }

  switch (pChannelExitParms->ExitId)
  {
      case MQXT_CHANNEL_SEC_EXIT:
         Traceit( "  MQCXP ExitId: MQXT_CHANNEL_SEC_EXIT\n");
         break;
      case MQXT_CHANNEL_MSG_EXIT:
         Traceit( "  MQCXP ExitId: MQXT_CHANNEL_MSG_EXIT\n");
         break;
      case MQXT_CHANNEL_SEND_EXIT:
         Traceit( "  MQCXP ExitId: MQXT_CHANNEL_SEND_EXIT\n");
         break;
      case MQXT_CHANNEL_RCV_EXIT:
         Traceit( "  MQCXP ExitId: MQXT_CHANNEL_RCV_EXIT\n");
         break;
      default:
         Traceit( "  *** Found unknown MQCXP ExitId: %ld\n",
                 pChannelExitParms->ExitId);
  } /* endswitch */

  switch (pChannelExitParms->ExitReason)
  {
      case MQXR_INIT:
         Traceit( "  MQCXP ExitReason: MQXR_INIT\n");
         break;
      case MQXR_TERM:
         Traceit( "  MQCXP ExitReason: MQXR_TERM\n");
         break;
      case MQXR_MSG:
         Traceit( "  MQCXP ExitReason: MQXR_MSG\n");
         break;
      case MQXR_XMIT:
         Traceit( "  MQCXP ExitReason: MQXR_XMIT\n");
         break;
      case MQXR_SEC_MSG:
         Traceit( "  MQCXP ExitReason: MQXR_SEC_MSG\n");
         break;
      case MQXR_INIT_SEC:
         Traceit(
                 "  MQCXP ExitReason: MQXR_INIT_SEC\n");
         break;
      default:
         Traceit( "  *** Found unknown MQCXP ExitReason: %ld\n",
                 pChannelExitParms->ExitReason);
  } /* endswitch */

  /* Check response field */

  if (pChannelExitParms->ExitResponse != MQXCC_OK)
  {
     Traceit( "  *** Found MQCXP ExitResponse: %ld\n",
             pChannelExitParms->ExitResponse);
  }

  /* Check response2 field */
  if (pChannelExitParms->ExitResponse2)
  {
     Traceit( "  *** Found MQCXP ExitResponse2: %ld\n",
             pChannelExitParms->ExitResponse2);
  }

  /* Check feedback field */
  if (pChannelExitParms->Feedback)
  {
     Traceit( "  *** Found MQCXP Feedback: %ld\n",
             pChannelExitParms->Feedback);
  }

  /* Print max segment length */
  if (pChannelExitParms->ExitReason != MQXR_INIT)
  {
     Traceit( "  MQCXP MaxSegmentLength: %ld\n",
             pChannelExitParms->MaxSegmentLength);
  }

  /* Print exit user area */
  Traceit( "  MQCXP ExitUserArea: %.16s\n",
          pChannelExitParms->ExitUserArea);

  Traceit("  MQCXP ExitUserArea(hex): ");
  for (p = pChannelExitParms->ExitUserArea, i=0;
       i < MQ_EXIT_USER_AREA_LENGTH ;
       i++)
  {
     Traceit("%02x ",*p++);
  } /* endfor */
  Traceit("\n");

  /* Print exit data */
  Traceit( "  MQCXP ExitData: %.32s\n",
          pChannelExitParms->ExitData);

  Traceit("  MQCXP ExitData(hex): ");
  for (p = pChannelExitParms->ExitData, i=0; i < MQ_EXIT_DATA_LENGTH ; i++ )
  {
     Traceit("%02x ",*p++);
  } /* endfor */
  Traceit("\n");
}



//
// -------------------- Dump the message data ---------------------------
//


void DumpMsgData (PMQXQH pMsg)
{
   int i;
   unsigned char *p;

  /* Check struc id */
  if (strncmp(pMsg->StrucId,
              MQXQH_STRUC_ID,
              sizeof(MQCHAR4)
             )
     )
     {
      Traceit( "  *** Found incorrect MQXQH Struc id: %.s\n",
             pMsg->StrucId);
     }

  /* Check version */
  if (pMsg->Version != MQXQH_VERSION_1)
  {
     Traceit( "  *** Found incorrect MQXQH Version: %ld\n",
             pMsg->Version);
  }

  /* Print RemoteQName */
  Traceit( "  MQXQH RemoteQName: %.48s\n",
          pMsg->RemoteQName);

  /* Print RemoteQMgrName */
  Traceit( "  MQXQH RemoteQMgrName: %.48s\n",
          pMsg->RemoteQMgrName);

  /* Print Msg Descriptor */
  Traceit( "  MQXQH MsgDesc data: \n");

  /* Print MsgDesc StrucId */
  Traceit( "    MQMD StrucId: %.4s\n",
          pMsg->MsgDesc.StrucId);

  /* Print MsgDesc Version */
  Traceit( "    MQMD StrucId: %ld\n",
          pMsg->MsgDesc.Version);

  /* Print MsgDesc Report options */
  Traceit( "    MQMD Report: %ld\n",
          pMsg->MsgDesc.Report);

  /* Print MsgDesc MsgType */
  Traceit( "    MQMD MsgType: %ld\n",
          pMsg->MsgDesc.MsgType);

  /* Print MsgDesc Expiry */
  Traceit( "    MQMD Expiry: %ld\n",
          pMsg->MsgDesc.Expiry);

  /* Print MsgDesc Feedback */
  Traceit( "    MQMD Feedback: %ld\n",
          pMsg->MsgDesc.Feedback);

  /* Print MsgDesc Encoding */
  Traceit( "    MQMD Encoding: %ld\n",
          pMsg->MsgDesc.Encoding);

  /* Print MsgDesc Coded Char Set id */
  Traceit( "    MQMD CodedCharSetId: %ld\n",
          pMsg->MsgDesc.CodedCharSetId);

  /* Print MsgDesc format  */
  Traceit( "    MQMD Format: %.48s\n",
          pMsg->MsgDesc.Format);

  /* Print MsgDesc Priority */
  Traceit( "    MQMD Priority: %ld\n",
          pMsg->MsgDesc.Priority);

  /* Print MsgDesc Persistence */
  Traceit(
          "    MQMD Persistence: %ld\n",
          pMsg->MsgDesc.Persistence);

  Traceit( "    MQMD MsgId: %.24s \n",
          pMsg->MsgDesc.MsgId);
  Traceit("    MQMD MsgId(hex): ");

  for (p = pMsg->MsgDesc.MsgId, i=0; i < 24 ; i++ )
  {
     Traceit("%02x ",*p++);
  } /* endfor */

  Traceit("\n");

  Traceit( "    MQMD CorrelId: %.24s \n",
          pMsg->MsgDesc.CorrelId);
  Traceit( "    MQMD CorrelId(hex): ");

  for (p = pMsg->MsgDesc.CorrelId, i=0; i < 24 ; i++ )
  {
     Traceit("%02x ",*p++);
  } /* endfor */

  Traceit("\n");

  Traceit( "    MQMD BackoutCount: %ld\n",
          pMsg->MsgDesc.BackoutCount);

  Traceit( "    MQMD ReplyToQ: %.48s\n",
          pMsg->MsgDesc.ReplyToQ);

  Traceit( "    MQMD ReplyToQMgr: %.48s\n",
          pMsg->MsgDesc.ReplyToQMgr);

  Traceit( "    MQMD UserIdentifier: %.12s\n",
          pMsg->MsgDesc.UserIdentifier);

  Traceit( "    MQMD AccountingToken: %.32s\n",
          pMsg->MsgDesc.AccountingToken);

  Traceit( "    MQMD ApplIdentityData: %.32s\n",
          pMsg->MsgDesc.ApplIdentityData);

  Traceit( "    MQMD PutApplType: %ld\n",
          pMsg->MsgDesc.PutApplType);

  Traceit( "    MQMD PutApplName: %.28s\n",
          pMsg->MsgDesc.PutApplName);

  Traceit( "    MQMD PutDate: %.8s\n",
          pMsg->MsgDesc.PutDate);

  Traceit( "    MQMD PutTime: %.8s\n",
          pMsg->MsgDesc.PutTime);

  Traceit( "    MQMD ApplOriginData: %.4s\n",
          pMsg->MsgDesc.ApplOriginData);
}


//
// -------------------- Channel Security Exit (what's really needed)-----
//


void MQENTRY SecurityExit (MQCXP        *pChannelExitParms,
                           MQCD         *pChannelDefinition,
                           MQLONG       *pDataLength,
                           MQLONG       *pAgentBufferLength,
                           MQBYTE       *pAgentBuffer,
                           MQLONG       *pExitBufferLength,
                           MQPTR        *pExitBufferAddr)
{
  int  iWordLen1=0,
       iWordLen2=0;


#ifdef USE_COMPUTER_NAME
  int  iConnNameLength;
  char vszConnName[MQ_CHANNEL_NAME_LENGTH+1];
#endif

  char *pszSecResp,
       *pszWord1,
       vszWorkBuffer[MQ_EXIT_DATA_LENGTH+1];

  MQLONG  *pP;

  OpenTraceFile();

  Traceit( "into Security exit for channel: %.20s\n",
          pChannelDefinition->ChannelName);

  CheckMQCXP(pChannelExitParms);

  //--------------------------//
  // Parse the security data  //
  // We should have 2 words   //
  //--------------------------//


  memcpy(&vszWorkBuffer[0],
         &pChannelExitParms->ExitData[0],
         32);

  vszWorkBuffer[MQ_EXIT_DATA_LENGTH]='\0'; // Ensure it is null terminated

  pszWord1    = strtok(vszWorkBuffer," ");  // Find first word

  if(pszWord1!=NULL)
  {
   iWordLen1   = strlen(pszWord1);
   pszSecResp  = strtok(NULL," ");  // Find next  word
   if(pszSecResp!=NULL)
   {
   iWordLen2   = strlen(pszSecResp);
   }
  }

  //----------------------------------//
  //  Trace data that's informative ! //
  //  remove this as required         //
  //----------------------------------//

  Traceit("  --> Queue Manager Names    (%.20s)\n",pChannelDefinition->QMgrName);
  Traceit("  --> Userid        is       (%.12s)\n",pChannelDefinition->UserIdentifier);
  Traceit("  --> Remote UserID is       (%.12s)\n",pChannelDefinition->RemoteUserIdentifier);
  Traceit("  --> McaName is             (%.20s)\n",pChannelDefinition->MCAName);
  Traceit("  --> ShortConnectionName is (%.20s)\n",pChannelDefinition->ShortConnectionName);
  Traceit("  --> Connection names is    (%.20s)\n",pChannelDefinition -> ConnectionName);
  Traceit("  --> Channel Name is        (%.20s)\n",pChannelDefinition -> ChannelName);
  Traceit("  --> AgentBuffer is         (%.32s)\n",pAgentBuffer);
  Traceit("  --> DataLength             (%ld)\n",*pDataLength);
  Traceit("  --> AgentBuffer length     (%ld)\n",*pAgentBufferLength);
  if(iWordLen1)
   Traceit("  --> SecResp     is         (%.32s)\n",pszSecResp);

 // ------------------- End Trace data ------------------------------------------

  switch (pChannelExitParms->ExitReason)
  {

       case MQXR_INIT:
       break;

       case MQXR_INIT_SEC:

          // Here when starting channel is at remote (host) end

          // Must have 2 words in the security exit data field
          // or none for this  example to work !

          if (iWordLen1)
          {

            // If security was requested for this channel send out our
            // security message to which the other end has to respond

            strncpy (pAgentBuffer, pChannelExitParms->ExitData,iWordLen1);
            *pDataLength = iWordLen1;

            if (iWordLen2)
               // This code expects to have 2 words of Security data
               // The challenge word and the Expected Response word.
            {
              Traceit("  --> Sending security request (%.20s)\n",pAgentBuffer);
              pChannelExitParms->ExitResponse = MQXCC_SEND_AND_REQUEST_SEC_MSG;
            }
          }
          else
          {
            *pDataLength=1;          // Must have a length to send else error
             pAgentBuffer[0]='!';    // Send something !
            Traceit("  --> Sending security request with NO DATA\n");
            pChannelExitParms->ExitResponse = MQXCC_SEND_AND_REQUEST_SEC_MSG;
          }

       break;


       case MQXR_SEC_MSG:        // Security message received
       {

         // If
         //    either  no data has been sent with the request
         //    or      The first word matches the 1st word in the security exit data
         //
         //      Then we Respond with the second word if any
         //      and add the Qmanager name to the response data.


         if ( (*pDataLength == 1 ) || ( iWordLen1 &&
                !strncmp(pAgentBuffer, pChannelExitParms->ExitData,iWordLen1)) )
         {
           // We have been sent a security message that may require a response */
           // So we send back the second word in the Security Exit Data
           //
           memset(pAgentBuffer,' ',32);                   //Set some of buffer
           *(pAgentBuffer+MQ_EXIT_DATA_LENGTH)='\0';      //Null terminate

           if (iWordLen2)   // If there is a second word in security exit data
           {
             memcpy (pAgentBuffer, pszSecResp, iWordLen2);
             *pDataLength = iWordLen2;
             pP += iWordLen2;

#ifdef USE_COMPUTER_NAME

           //----------------------------------------------------------------//
           //  Add Computer name after this word if is is defined.           //
           // This method of identifying this computer to the remote end     //
           //  works , but you may have problems if you don't use unique     //
           //  Queue Manager names.as the message ID is partly made up with  //
           //  the first 12 characters on QMgrName and a time stamp          //
           //  See the MQ Application Programming Reference P43              //
           //----------------------------------------------------------------//
              vszConnName[0]='\0';
              iConnNameLength = GetProfileString( MQWAPPNAME, MQWCONNNAME, "",
                                vszConnName, sizeof(vszConnName) );
             if(iConnNameLength)
             {
              memcpy (pAgentBuffer+iWordLen2+1,&vszConnName[0],iConnNameLength+1);

              *pDataLength+= iConnNameLength+1;
             }
#else
           //----------------------------------------------------------------//
           //  Normally we recommend you send  the QManager Name (see above) //
           //  This means we MUST have unique Qmanager Names of EVERY machine//
           //  Currently the first 12 of 48 possible  characters are used    //
           //  on all platforms , so that's what we do here                  //
           //  It's your choice !                                            //
           //----------------------------------------------------------------//

              *(pAgentBuffer+iWordLen2)=' ';
              memcpy (pAgentBuffer+iWordLen2+1,pChannelDefinition->QMgrName,12);
              *pDataLength+= 12+1;

#endif
           }
           else      // No security data to append to, just send id
           {
#ifdef USE_COMPUTER_NAME

           //
           //  add Computer name after this word if is is defined.
           //
              vszConnName[0]='\0';
              iConnNameLength = GetProfileString( MQWAPPNAME, MQWCONNNAME, "",
                                vszConnName, sizeof(vszConnName) );
             if(iConnNameLength)
             {
              memcpy (pAgentBuffer,&vszConnName[0],iConnNameLength);

              *pDataLength+= iConnNameLength;
             }
#else
           //
           //  Normally we should use our QManager to send to remote channel
           //  This means we MUST have unique Qmanager Names of EVERY machine
           //  Currently the first 12 of 48 possible  characters are used
           //


              memcpy (pAgentBuffer,pChannelDefinition->QMgrName,12);
              *pDataLength+= 12;
#endif
           }


             Traceit("  --> Sending security response (%.32s)\n",pAgentBuffer);
             Traceit("  --> Security response datalength (%ld)\n",*pDataLength);
             pChannelExitParms->ExitResponse = MQXCC_SEND_SEC_MSG;

         }
                //-----------------------------//
                // Here if on Remote system    //
                //-----------------------------//
         else
         {

          if(iWordLen2)  // if security response  expected
          {
            if (strncmp(pAgentBuffer, pszSecResp, iWordLen2))
            {

             // We have received a security message but it is not an expected
             // response to our message (if we sent one

             Traceit("  --> Received invalid security msg(%.32s)\n",pszSecResp);
             Traceit("  --> Strncmp(%s,%s,%d)",pAgentBuffer,pszSecResp,iWordLen2);
             pChannelExitParms->ExitResponse = MQXCC_SUPPRESS_FUNCTION;
             break;
            }
            else
            {

             Traceit("  --> Received valid security response\n");
             // Here we want to see if we need to change our channel name
             // to that sent by requestor, as it may be a slip/ppp DHCP id
             //
             // Check if we have another word in the AgentBuffer
             //
             pszWord1 = strtok(pAgentBuffer," ");  // Find first word
             pszWord1 = strtok(NULL," ");          // Find Second Word
            }
          }
          else
          {
             pszWord1 = strtok(pAgentBuffer," ");  // Find first word
          }



          //---------------------------------------------------------------//
          // we always expect 1 word of unique identifier back in response //
          //---------------------------------------------------------------//

          if (pszWord1 != NULL)
          {
           Traceit("  --> Received Unique Identifier (%s)\n",pszWord1);

           memset(pChannelDefinition->ConnectionName,' ',(size_t)sizeof(pChannelDefinition->ConnectionName) );
           memcpy(pChannelDefinition ->ConnectionName,pszWord1,strlen(pszWord1));

           Traceit("  --> Changed Connection Name to (%s)\n",
                                 pChannelDefinition ->ConnectionName);
          }
          //----------------------------------------------------//
          // All ok, so message transfer should proceed normally//
          //----------------------------------------------------//

             pChannelExitParms->ExitResponse = MQXCC_OK;

         } // end else on remote system
       }
       break;

       case MQXR_TERM:          // MCA Termination , free your memory !
        break;
       default:
       break;

  } /* endswitch */

  Traceit( "Leaving Security exit: %ld \n\n",pChannelExitParms->ExitResponse);
  if (stream!=NULL) fclose( stream );
}


//
//---------------------  Sample Message Exit --------------------
//

void MQENTRY MessageExit  (MQCXP        *pChannelExitParms,
                           MQCD         *pChannelDefinition,
                           MQLONG       *pDataLength,
                           MQLONG       *pAgentBufferLength,
                           MQBYTE       *pAgentBuffer,
                           MQLONG       *pExitBufferLength,
                           MQPTR        *pExitBufferAddr)
{
  OpenTraceFile();
  Traceit( "into Message exit for channel: %.20s\n",
          pChannelDefinition->ChannelName);
  CheckMQCXP(pChannelExitParms);
  Traceit( "  DataLength: %ld,  AgentBufferLength: %ld\n",
          *pDataLength,
          *pAgentBufferLength);

  /* Examine the message descriptor to find information about the msg       */
  if (*pAgentBufferLength &&
      !strncmp(pChannelExitParms->ExitData,
               "dump",
               4)
     )
  {
    Traceit( "  Agent buffer at: %lx\n",
          pAgentBuffer);
    DumpMsgData((PMQXQH)pAgentBuffer);
  }

  /* If channel data indicates discard all messages */
  if (pChannelExitParms->ExitReason == MQXR_MSG &&
      !strncmp(pChannelExitParms->ExitData,
               "discard",
               7)
     )
  {
    Traceit( "  Discarding message\n");
    pChannelExitParms->ExitResponse = MQXCC_SUPPRESS_FUNCTION;
  }

  /* If message data indicates stop channel */
  if (pChannelExitParms->ExitReason == MQXR_MSG &&
      !strncmp((pAgentBuffer+sizeof(MQXQH)),
               "STOP CHANNEL",
               12)
     )
  {
    Traceit( "  STOPPING CHANNEL !! \n");
    pChannelExitParms->ExitResponse = MQXCC_CLOSE_CHANNEL;
  }
  Traceit( "leaving Message exit: %ld \n\n",
          pChannelExitParms->ExitResponse);
  if (stream!=NULL) fclose( stream );
}


//
//---------------------  Sample Send Exit --------------------
//



void MQENTRY SendExit     (MQCXP        *pChannelExitParms,
                           MQCD         *pChannelDefinition,
                           MQLONG       *pDataLength,
                           MQLONG       *pAgentBufferLength,
                           MQBYTE       *pAgentBuffer,
                           MQLONG       *pExitBufferLength,
                           MQPTR        *pExitBufferAddr)
{
  OpenTraceFile();
  Traceit( "into Send exit for channel: %.20s\n",
          pChannelDefinition->ChannelName);;
  CheckMQCXP(pChannelExitParms);
  Traceit( "  DataLength: %ld,  AgentBufferLength: %ld\n",
          *pDataLength,
          *pAgentBufferLength);
  Traceit( "leaving Send exit: %ld \n\n",
          pChannelExitParms->ExitResponse);
  if (stream!=NULL) fclose( stream );
}



//
//---------------------  Sample Receive Exit --------------------
//



void MQENTRY ReceiveExit  (MQCXP        *pChannelExitParms,
                           MQCD         *pChannelDefinition,
                           MQLONG       *pDataLength,
                           MQLONG       *pAgentBufferLength,
                           MQBYTE       *pAgentBuffer,
                           MQLONG       *pExitBufferLength,
                           MQPTR        *pExitBufferAddr)
{
  OpenTraceFile();
  Traceit( "into Receive exit for channel: %.20s\n",
          pChannelDefinition->ChannelName);
  CheckMQCXP(pChannelExitParms);
  Traceit( "  DataLength: %ld,  AgentBufferLength: %ld\n",
          *pDataLength,
          *pAgentBufferLength);
  Traceit( "leaving Receive exit: %ld \n\n",
          pChannelExitParms->ExitResponse);

  if (stream!=NULL) fclose( stream );
}

int FAR PASCAL _WEP(int WEPParam)
{
       /* Your WEP functionality goes here */
  return 1;
}

